#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""
Encoder model wrappers based on HuggingFace code
"""

import logging
from typing import Tuple

import torch
from torch import Tensor as T
from torch import nn
from transformers.optimization import AdamW
# from transformers.modeling_bert import BertConfig, BertModel
# from transformers.tokenization_bert import BertTokenizer
# from transformers.tokenization_roberta import RobertaTokenizer
from transformers import BertConfig, BertModel, BertTokenizer, RobertaTokenizer

from dpr.utils.data_utils import Tensorizer
from .biencoder import BiEncoder, SharedBiEncoder, HybridModel
from .reader import Reader

logger = logging.getLogger(__name__)


def get_bert_biencoder_components(args, inference_only: bool = False, **kwargs):
    dropout = args.dropout if hasattr(args, "dropout") else 0.0

    ctx_encoder = HFBertEncoder.init_encoder(
        args.pretrained_model_cfg,
        projection_dim=args.projection_dim,
        dropout=dropout,
        **kwargs
    )

    fix_ctx_encoder = (
        args.fix_ctx_encoder if hasattr(args, "fix_ctx_encoder") else False
    )
    return_pooled_ctx_repr = (
        args.scoring == "dpr" if hasattr(args, "scoring") else True
    )
    if not args.weight_sharing:
        question_encoder = HFBertEncoder.init_encoder(
            args.pretrained_model_cfg,
            projection_dim=args.projection_dim,
            dropout=dropout,
            **kwargs
        )
        fix_question_encoder = (
            args.fix_question_encoder if hasattr(args, "fix_question_encoder") else False
        )
        biencoder = BiEncoder(
            question_encoder, ctx_encoder,
            fix_ctx_encoder=fix_ctx_encoder,
            fix_q_encoder=fix_question_encoder,
            return_pooled_ctx_repr=return_pooled_ctx_repr
        )
    else:
        biencoder = SharedBiEncoder(
            ctx_encoder, fix_ctx_encoder, return_pooled_ctx_repr
        )

    if args.hybrid_dense_sparse:
        biencoder = HybridModel(biencoder,
                                (args.hybrid_model_alpha_init if hasattr(args, "hybrid_model_alpha_init") else 1.0))

    optimizer = (
        get_optimizer(
            biencoder,
            learning_rate=args.learning_rate,
            adam_eps=args.adam_eps,
            weight_decay=args.weight_decay,
        )
        if not inference_only
        else None
    )

    tensorizer = get_bert_tensorizer(args)

    return tensorizer, biencoder, optimizer


def get_bert_reader_components(args, inference_only: bool = False, **kwargs):
    dropout = args.dropout if hasattr(args, "dropout") else 0.0
    encoder = HFBertEncoder.init_encoder(
        args.pretrained_model_cfg, projection_dim=args.projection_dim, dropout=dropout
    )

    hidden_size = encoder.config.hidden_size
    reader = Reader(encoder, hidden_size)

    optimizer = (
        get_optimizer(
            reader,
            learning_rate=args.learning_rate,
            adam_eps=args.adam_eps,
            weight_decay=args.weight_decay,
        )
        if not inference_only
        else None
    )

    tensorizer = get_bert_tensorizer(args)
    return tensorizer, reader, optimizer


def get_bert_tensorizer(args, tokenizer=None):
    if not tokenizer:
        tokenizer = get_bert_tokenizer(
            args.pretrained_model_cfg, do_lower_case=args.do_lower_case
        )
    return BertTensorizer(tokenizer, args.sequence_length)


def get_roberta_tensorizer(args, tokenizer=None):
    if not tokenizer:
        tokenizer = get_roberta_tokenizer(
            args.pretrained_model_cfg, do_lower_case=args.do_lower_case
        )
    return RobertaTensorizer(tokenizer, args.sequence_length)


def get_optimizer(
    model: nn.Module,
    learning_rate: float = 1e-5,
    adam_eps: float = 1e-8,
    weight_decay: float = 0.0,
) -> torch.optim.Optimizer:
    no_decay = ["bias", "LayerNorm.weight"]

    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": weight_decay,
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]
    optimizer = AdamW(optimizer_grouped_parameters, lr=learning_rate, eps=adam_eps)
    return optimizer


def get_bert_tokenizer(pretrained_cfg_name: str, do_lower_case: bool = True):
    if pretrained_cfg_name:
        return BertTokenizer.from_pretrained(
            pretrained_cfg_name, do_lower_case=do_lower_case
        )
    else:
        return BertTokenizer.from_pretrained("bert-base-uncased" if do_lower_case else "bert-base-cased",
                                             do_lower_case=do_lower_case)


def get_roberta_tokenizer(pretrained_cfg_name: str, do_lower_case: bool = True):
    # still uses HF code for tokenizer since they are the same
    return RobertaTokenizer.from_pretrained(
        pretrained_cfg_name, do_lower_case=do_lower_case
    )


class HFBertEncoder(BertModel):
    def __init__(self, config, project_dim: int = 0):
        BertModel.__init__(self, config)
        assert config.hidden_size > 0, "Encoder hidden_size can't be zero"
        self.encode_proj = (
            nn.Linear(config.hidden_size, project_dim) if project_dim != 0 else None
        )
        self.init_weights()

    @classmethod
    def init_encoder(
        cls, cfg_name: str, projection_dim: int = 0, dropout: float = 0.1, **kwargs
    ) -> BertModel:
        cfg = BertConfig.from_pretrained(cfg_name if cfg_name else "bert-base-uncased")
        if dropout != 0:
            cfg.attention_probs_dropout_prob = dropout
            cfg.hidden_dropout_prob = dropout
        if cfg_name:
            return cls.from_pretrained(cfg_name, config=cfg, project_dim=projection_dim, **kwargs)
        else:
            return cls(config=cfg, project_dim=projection_dim)

    def forward(
        self, input_ids: T, token_type_ids: T, attention_mask: T
    ) -> Tuple[T, ...]:
        res = super().forward(
            input_ids=input_ids,
            token_type_ids=token_type_ids,
            attention_mask=attention_mask,
        )

        if isinstance(res, dict):
            # Transformers 4
            sequence_output = res["last_hidden_state"]
            hidden_states = res["hidden_states"] if "hidden_states" in res else None
        else:
            # Transformers 3
            assert isinstance(res, tuple)
            sequence_output, _, *rest = res
            hidden_states = rest[0] if rest else None

        pooled_output = sequence_output[:, 0, :]
        if self.encode_proj:
            pooled_output = self.encode_proj(pooled_output)
        return sequence_output, pooled_output, hidden_states

    def get_out_size(self):
        if self.encode_proj:
            return self.encode_proj.out_features
        return self.config.hidden_size


class BertTensorizer(Tensorizer):
    def __init__(
        self, tokenizer: BertTokenizer, max_length: int, pad_to_max: bool = True
    ):
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.pad_to_max = pad_to_max

    def text_to_tensor(
        self, text, title=None, add_special_tokens=True, max_length=None
    ):
        if isinstance(text, str):
            text = text.strip()

        seq_len = self.max_length if max_length is None else max_length
        # tokenizer automatic padding is explicitly disabled since its inconsistent behavior
        if title:
            token_ids = self.tokenizer.encode(
                title,
                text_pair=text,
                add_special_tokens=add_special_tokens,
                max_length=seq_len,
                pad_to_max_length=False,
                truncation=True,
            )
        else:
            token_ids = self.tokenizer.encode(
                text,
                add_special_tokens=add_special_tokens,
                max_length=seq_len,
                pad_to_max_length=False,
                truncation=True,
            )

        if self.pad_to_max and len(token_ids) < seq_len:
            token_ids = token_ids + [self.tokenizer.pad_token_id] * (
                seq_len - len(token_ids)
            )
        if len(token_ids) > seq_len:
            token_ids = token_ids[0:seq_len]
            token_ids[-1] = self.tokenizer.sep_token_id

        return torch.tensor(token_ids)

    def get_pair_separator_ids(self) -> T:
        return torch.tensor([self.tokenizer.sep_token_id])

    def get_pad_id(self) -> int:
        return self.tokenizer.pad_token_id

    def get_attn_mask(self, tokens_tensor: T) -> T:
        return tokens_tensor != self.get_pad_id()

    def is_sub_word_id(self, token_id: int):
        token = self.tokenizer.convert_ids_to_tokens([token_id])[0]
        return token.startswith("##") or token.startswith(" ##")

    def to_string(self, token_ids, skip_special_tokens=True):
        return self.tokenizer.decode(token_ids, skip_special_tokens=True)

    def set_pad_to_max(self, do_pad: bool):
        self.pad_to_max = do_pad


class RobertaTensorizer(BertTensorizer):
    def __init__(self, tokenizer, max_length: int, pad_to_max: bool = True):
        super(RobertaTensorizer, self).__init__(
            tokenizer, max_length, pad_to_max=pad_to_max
        )
